import os

from watchdog.events import FileSystemEventHandler
from watchdog.utils import has_attribute
from watchdog.utils import unicode_paths


class VcsEventHandler(FileSystemEventHandler):

    def __init__(self, vcs):
        super(VcsEventHandler, self).__init__()
        self.vcs = vcs

    def dispatch(self, event):
        """Only dispatch if the event does not correspond to an ignored file.
        Args:
            event (watchdog.events.FileSystemEvent)
        """
        if event.is_directory:
            return
        paths = []
        if has_attribute(event, 'dest_path'):
            paths.append(os.path.realpath(
                unicode_paths.decode(event.dest_path)))
        if event.src_path:
            paths.append(os.path.realpath(
                unicode_paths.decode(event.src_path)))
        paths = [p for p in paths
                 if not p.startswith(os.path.realpath(self.vcs.repository_dir()))
                 and not self.vcs.path_is_ignored(p)]

        if len(paths) > 0:
            super(VcsEventHandler, self).dispatch(event)



import click

from watchdog.observers import Observer

from easyci.file_system_events.tests_event_handler import TestsEventHandler


@click.command()
@click.pass_context
def watch(ctx):
    """Watch the directory for changes. Automatically run tests.
    """
    vcs = ctx.obj['vcs']

    event_handler = TestsEventHandler(vcs)

    observer = Observer()
    observer.schedule(event_handler, vcs.path, recursive=True)
    observer.start()
    click.echo('Watching directory `{path}`. Use ctrl-c to stop.'.format(path=vcs.path))
    while observer.isAlive():
        observer.join(timeout=1)



import filelock
import os

from filelock import Timeout

from contextlib import contextmanager
from enum import Enum, unique


# Timeout is imported so it can be imported by other modules as a member
# of this module, not filelock
Timeout


@unique
class Lock(Enum):
    tests_history = 1


def _get_lock_path(vcs, lock_object):
    return os.path.join(vcs.private_dir(), 'locks', lock_object.name)


def init(vcs):
    """Initialize the locking module for a repository
    """
    path = os.path.join(vcs.private_dir(), 'locks')
    if not os.path.exists(path):
        os.mkdir(path)


@contextmanager
def lock(vcs, lock_object, wait=True):
    """A context manager that grabs the lock and releases it when done.

    This blocks until the lock can be acquired.

    Args:
        vcs (easyci.vcs.base.Vcs)
        lock_object (Lock)
        wait (boolean) - whether to wait for the lock or error out

    Raises:
        Timeout
    """
    if wait:
        timeout = -1
    else:
        timeout = 0
    lock_path = _get_lock_path(vcs, lock_object)
    lock = filelock.FileLock(lock_path)
    with lock.acquire(timeout=timeout):
        yield



import os


class VersionNotInstalledError(Exception):
    pass


def _get_version_path(vcs):
    return os.path.join(vcs.private_dir(), 'version')


def get_installed_version(vcs):
    """Get the installed version for this project.

    Args:
        vcs (easyci.vcs.base.Vcs)

    Returns:
        str - version number

    Raises:
        VersionNotInstalledError
    """
    version_path = _get_version_path(vcs)
    if not os.path.exists(version_path):
        raise VersionNotInstalledError
    with open(version_path, 'r') as f:
        return f.read().strip()


def set_installed_version(vcs, version):
    """Set the installed version for this project.

    Args:
        vcs (easyci.vcs.base.Vcs)
        version (str)
    """
    version_path = _get_version_path(vcs)
    with open(version_path, 'w') as f:
        f.write(version)



import os
import shutil
import subprocess32 as subprocess


class ResultsNotFoundError(Exception):
    pass


def _get_results_directory(vcs, signature):
    return os.path.join(vcs.private_dir(), 'results', signature)


def save_results(vcs, signature, result_path, patterns):
    """Save results matching `patterns` at `result_path`.

    Args:
        vcs (easyci.vcs.base.Vcs) - the VCS object for the actual project
                                    (not the disposable copy)
        signature (str) - the project state signature
        result_path (str) - the path containing the result, usually
                            a disposable copy of the project
        patterns (str) - `rsync`-compatible patterns matching test results
                         to save.
    """
    results_directory = _get_results_directory(vcs, signature)
    if not os.path.exists(results_directory):
        os.makedirs(results_directory)
    with open(os.path.join(results_directory, 'patterns'), 'w') as f:
        f.write('\n'.join(patterns))
    if not os.path.exists(os.path.join(results_directory, 'results')):
        os.mkdir(os.path.join(results_directory, 'results'))
    includes = ['--include={}'.format(x)
                for x in patterns]
    cmd = ['rsync', '-r'] + includes + ['--exclude=*',
                                        os.path.join(result_path, ''),
                                        os.path.join(results_directory, 'results', '')]
    subprocess.check_call(cmd)


def sync_results(vcs, signature):
    """Sync the saved results for `signature` back to the project.

    Args:
        vcs (easyci.vcs.base.Vcs)
        signature (str)
    Raises:
        ResultsNotFoundError
    """
    results_directory = _get_results_directory(vcs, signature)
    if not os.path.exists(results_directory):
        raise ResultsNotFoundError
    with open(os.path.join(results_directory, 'patterns'), 'r') as f:
        patterns = f.read().strip().split()
    includes = ['--include={}'.format(x)
                for x in patterns]
    cmd = ['rsync', '-r'] + includes + ['--exclude=*',
                                        os.path.join(
                                            results_directory, 'results', ''),
                                        os.path.join(vcs.path, '')]
    subprocess.check_call(cmd)


def remove_results(vcs, signature):
    """Removed saved results for this signature

    Args:
        vcs (easyci.vcs.base.Vcs)
        signature (str)
    Raises:
        ResultsNotFoundError
    """
    results_directory = _get_results_directory(vcs, signature)
    if not os.path.exists(results_directory):
        raise ResultsNotFoundError
    shutil.rmtree(results_directory)


def get_signatures_with_results(vcs):
    """Returns the list of signatures for which test results are saved.

    Args:
        vcs (easyci.vcs.base.Vcs)

    Returns:
        List[str]
    """
    results_dir = os.path.join(vcs.private_dir(), 'results')
    if not os.path.exists(results_dir):
        return []
    rel_paths = os.listdir(results_dir)
    return [p for p in rel_paths if os.path.isdir(os.path.join(results_dir, p))]



import click
import subprocess32 as subprocess

from easyci import exit_codes, locking
from easyci.history import (
    commit_signature, get_committed_signatures, get_staged_signatures,
    stage_signature, unstage_signature,
)
from easyci.results import save_results, sync_results, ResultsNotFoundError
from easyci.utils import contextmanagers
from easyci.user_config import (
    load_user_config, ConfigFormatError, ConfigNotFoundError,
    _default_config
)


@click.command()
@click.option('--staged-only', '-s', is_flag=True, default=False, help='Test against staged version of the repo. Remove all unstaged and ignored files before running.')
@click.option('--head-only', '-h', is_flag=True, default=False, help='Test against the current HEAD. Resets to HEAD and remove all ignored files before running.')
@click.pass_context
def test(ctx, staged_only, head_only):
    """Run tests. If a passing test run is found in the tests run history,
    then this does not run any tests.
    """
    git = ctx.obj['vcs']

    click.echo('Making a temporary copy of your project...', nl=False)
    with git.temp_copy() as copy:
        click.echo('Done.')
        if head_only:
            click.echo('Resetting to HEAD...', nl=False)
            copy.clear('HEAD')
            click.echo('Done.')
        elif staged_only:
            click.echo('Resetting to staged files...', nl=False)
            copy.remove_unstaged_files()
            click.echo('Done.')

        click.echo('Loading config file...', nl=False)
        try:
            config = load_user_config(copy)
            click.echo('Done.')
        except ConfigFormatError:
            click.echo("Invalid config")
            ctx.abort()
        except ConfigNotFoundError:
            click.echo("No config file")
            config = _default_config

        click.echo('Checking if tests were already ran...', nl=False)
        new_signature = copy.get_signature()

        with locking.lock(git, locking.Lock.tests_history):
            in_committed = new_signature in get_committed_signatures(git)
            in_staged = new_signature in get_staged_signatures(git)
            if not in_committed and not in_staged:
                stage_signature(git, new_signature)
            if in_committed:
                click.echo('')
                click.echo(click.style('OK', bg='green', fg='black') +
                           ' Tests already ran.')
                try:
                    click.echo('Syncing test results...', nl=False)
                    sync_results(git, new_signature)
                    click.echo('Done')
                except ResultsNotFoundError:
                    click.echo('No results to sync.')
                ctx.exit(exit_codes.SUCCESS)
            if in_staged:
                click.echo('')
                click.echo(click.style('In Progress', bg='yellow', fg='black') +
                           ' Tests already running.')
                ctx.exit(exit_codes.ALREADY_RUNNING)
        click.echo('Done.')
        with contextmanagers.chdir(copy.path):
            all_passed = True
            for test in config['tests']:
                click.echo('Running test: {}'.format(test))

                # ok to use shell=True, as the whole point of EasyCI is to run
                # arbitrary code
                ret = subprocess.call(test, shell=True)
                if ret == 0:
                    click.secho('Passed', bg='green', fg='black')
                else:
                    click.secho('Failed', bg='red', fg='black')
                    all_passed = False

        with locking.lock(git, locking.Lock.tests_history):
            # collect results
            if len(config['collect_results']) > 0:
                click.echo('Collecting results...', nl=False)
                save_results(git, new_signature, copy.path,
                             config['collect_results'])
                click.echo('Done')
                click.echo('Syncing test results...', nl=False)
                sync_results(git, new_signature)
                click.echo('Done')

            # save signature
            if not all_passed:
                unstage_signature(git, new_signature)
                ctx.exit(exit_codes.FAILURE)
            else:
                commit_signature(git, config, new_signature)
                ctx.exit(exit_codes.SUCCESS)



import os


class NotStagedError(Exception):
    pass


class AlreadyStagedError(Exception):
    pass


class AlreadyCommittedError(Exception):
    pass


def _get_committed_history_path(vcs):
    """Get the path to the file containing committed tests history

    Args:
        vcs (easyci.vcs.base.Vcs)

    Returns:
        str - the path
    """
    return os.path.join(vcs.private_dir(), 'passed')


def _get_staged_history_path(vcs):
    """Get the path to the file containing staged tests history

    Args:
        vcs (easyci.vcs.base.Vcs)

    Returns:
        str - the path
    """
    return os.path.join(vcs.private_dir(), 'in_progress')


def get_committed_signatures(vcs):
    """Get the list of committed signatures

    Args:
        vcs (easyci.vcs.base.Vcs)

    Returns:
        list(basestring) - list of signatures
    """
    committed_path = _get_committed_history_path(vcs)
    known_signatures = []
    if os.path.exists(committed_path):
        with open(committed_path, 'r') as f:
            known_signatures = f.read().split()
    return known_signatures


def get_staged_signatures(vcs):
    """Get the list of staged signatures

    Args:
        vcs (easyci.vcs.base.Vcs)

    Returns:
        list(basestring) - list of signatures
    """
    staged_path = _get_staged_history_path(vcs)
    known_signatures = []
    if os.path.exists(staged_path):
        with open(staged_path, 'r') as f:
            known_signatures = f.read().split()
    return known_signatures


def commit_signature(vcs, user_config, signature):
    """Add `signature` to the list of committed signatures

    The signature must already be staged

    Args:
        vcs (easyci.vcs.base.Vcs)
        user_config (dict)
        signature (basestring)

    Raises:
        NotStagedError
        AlreadyCommittedError
    """
    if signature not in get_staged_signatures(vcs):
        raise NotStagedError
    evidence_path = _get_committed_history_path(vcs)
    committed_signatures = get_committed_signatures(vcs)
    if signature in committed_signatures:
        raise AlreadyCommittedError
    committed_signatures.append(signature)
    string = '\n'.join(committed_signatures[-user_config['history_limit']:])
    with open(evidence_path, 'w') as f:
        f.write(string)
    unstage_signature(vcs, signature)


def stage_signature(vcs, signature):
    """Add `signature` to the list of staged signatures

    Args:
        vcs (easyci.vcs.base.Vcs)
        signature (basestring)

    Raises:
        AlreadyStagedError
    """
    evidence_path = _get_staged_history_path(vcs)
    staged = get_staged_signatures(vcs)
    if signature in staged:
        raise AlreadyStagedError
    staged.append(signature)
    string = '\n'.join(staged)
    with open(evidence_path, 'w') as f:
        f.write(string)


def unstage_signature(vcs, signature):
    """Remove `signature` from the list of staged signatures

    Args:
        vcs (easyci.vcs.base.Vcs)
        signature (basestring)

    Raises:
        NotStagedError
    """
    evidence_path = _get_staged_history_path(vcs)
    staged = get_staged_signatures(vcs)
    if signature not in staged:
        raise NotStagedError
    staged.remove(signature)
    string = '\n'.join(staged)
    with open(evidence_path, 'w') as f:
        f.write(string)


def clear_history(vcs):
    """Clear (committed) test run history from this project.

    Args:
        vcs (easyci.vcs.base.Vcs)
    """
    evidence_path = _get_committed_history_path(vcs)
    if os.path.exists(evidence_path):
        os.remove(evidence_path)



import os
import yaml


class ConfigFormatError(Exception):
    pass


class ConfigNotFoundError(Exception):
    pass


_default_config = {
    "tests": [],
    "collect_results": [],
    "history_limit": 100,
}


_config_types = {
    "tests": list,
    "collect_results": list,
    "history_limit": int,
}


def load_user_config(vcs):
    """Load the user config

    Args:
        vcs (easyci.vcs.base.Vcs) - the vcs object for the current project

    Returns:
        dict - the config

    Raises:
        ConfigFormatError
        ConfigNotFoundError
    """
    config_path = os.path.join(vcs.path, 'eci.yaml')
    if not os.path.exists(config_path):
        raise ConfigNotFoundError
    with open(config_path, 'r') as f:
        try:
            config = yaml.safe_load(f)
        except yaml.YAMLError:
            raise ConfigFormatError
    if not isinstance(config, dict):
        raise ConfigFormatError
    for k, v in _default_config.iteritems():
        config.setdefault(k, v)
    for k, v in _config_types.iteritems():
        if not isinstance(config[k], v):
            raise ConfigFormatError
    return config



import os
import shutil
import tempfile

from contextlib import contextmanager


@contextmanager
def chdir(path):
    """Change the working directory to `path` for the duration of this context
    manager.

    Args:
        path (str)
    """
    cur_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cur_cwd)


@contextmanager
def temp_dir():
    """Create a temporary folder for the duration of this context manager,
    deleting it afterwards.

    Yields:
        str - path to the temporary folder
    """
    path = tempfile.mkdtemp()
    try:
        yield path
    finally:
        shutil.rmtree(path)



import pkg_resources


class HookNotFoundError(Exception):
    pass


def get_hook(hook_name):
    """Returns the specified hook.

    Args:
        hook_name (str)

    Returns:
        str - (the content of) the hook

    Raises:
        HookNotFoundError
    """
    if not pkg_resources.resource_exists(__name__, hook_name):
        raise HookNotFoundError
    return pkg_resources.resource_string(__name__, hook_name)



import click
import os
import subprocess32 as subprocess
import watchdog.events
import watchdog.utils

from multiprocessing import dummy as multiprocessing

from easyci import exit_codes
from easyci.file_system_events.vcs_event_handler import VcsEventHandler


class TestsEventHandler(VcsEventHandler):

    _exit_messages = {
        exit_codes.SUCCESS: click.style('Passed', bg='green', fg='black'),
        exit_codes.FAILURE: click.style('Failed', bg='red', fg='black'),
        exit_codes.ALREADY_RUNNING: click.style('Already Running', bg='yellow', fg='black'),
    }

    def __init__(self, vcs):
        super(TestsEventHandler, self).__init__(vcs)
        self.pool = multiprocessing.Pool()

    def _get_relative_path(self, path):
        return os.path.relpath(path, self.vcs.path)

    def _get_event_string(self, event):
        src = self._get_relative_path(
            watchdog.utils.unicode_paths.decode(event.src_path))
        if event.event_type == watchdog.events.EVENT_TYPE_CREATED:
            return 'Created: ' + src
        elif event.event_type == watchdog.events.EVENT_TYPE_DELETED:
            return 'Deleted: ' + src
        elif event.event_type == watchdog.events.EVENT_TYPE_MODIFIED:
            return 'Modified: ' + src
        elif event.event_type == watchdog.events.EVENT_TYPE_MOVED:
            dest = self._get_relative_path(
                watchdog.utils.unicode_paths.decode(event.dest_path))
            return 'Moved: {src} to {dest}'.format(src=src, dest=dest)
        else:
            return str(event)

    def on_any_event(self, event):
        str_event = self._get_event_string(event)

        def callback(code):
            click.echo(self._exit_messages[code] + ' ' + str_event)
        click.echo(click.style('Detected', bg='yellow',
                               fg='black') + ' ' + str_event)
        devnull = open(os.devnull, 'w')
        self.pool.apply_async(subprocess.call, [['eci', 'test']], {
            'stdout': devnull,
            'stderr': devnull,
        }, callback=callback)



import click
import os
import yaml

import easyci

from easyci import locking
from easyci.hooks import hooks_manager
from easyci.version import set_installed_version


@click.command()
@click.pass_context
def init(ctx):
    """Initialize the project for use with EasyCI. This installs the necessary
    git hooks (pre-commit + pre-push) and add a config file if one does not
    already exists.
    """
    # install hooks
    git = ctx.obj['vcs']
    click.echo("Installing hooks...", nl=False)
    for old in ['commit-msg']:
        path = os.path.join(git.path, '.git/hooks', old)
        if os.path.exists(path):
            os.remove(path)
    for new in ['pre-commit', 'pre-push']:
        git.install_hook(new, hooks_manager.get_hook(new))
    click.echo("Done.")

    # add a config file if one does not exist
    config_path = os.path.join(git.path, 'eci.yaml')
    if not os.path.exists(config_path):
        click.echo("Placing a trivial config file in your project...", nl=False)
        with open(config_path, 'w') as f:
            f.write(yaml.safe_dump(
                {'tests': ['echo please modify to run your tests', 'true']}))
        click.echo("Done.")

    # initialize lock
    locking.init(git)

    # update installed version
    click.echo("Updating installed version...", nl=False)
    set_installed_version(git, easyci.__version__)
    click.echo("Done.")



"""
This module is inspired by the open-sourced project Changes:
https://github.com/dropbox/changes
"""
from __future__ import absolute_import

import os
import os.path

from contextlib import contextmanager
from subprocess32 import Popen, PIPE, check_call

from easyci.utils import contextmanagers


class CommandError(Exception):

    def __init__(self, cmd, retcode, stdout, stderr):
        self.cmd = cmd
        self.retcode = retcode
        self.stdout = stdout
        self.stderr = stderr

    def __unicode__(self):
        return '%s returned %d:\nSTDOUT: %r\nSTDERR: %r' % (
            self.cmd, self.retcode, self.stdout, self.stderr)  # pragma: no cover

    def __str__(self):
        return self.__unicode__().encode('utf-8')  # pragma: no cover


class Vcs(object):

    def __init__(self, path=None):
        """Initialize a new Vcs object for a repository located at `path`.
        If `path` is `None`, then `get_working_directory` is used to identify
        the path.

        Args:
            path (str) - optional. The path to the repo working directory.
        """
        self.path = None
        if path is None:
            self.path = self.get_working_directory()
        else:
            self.path = path
        assert self.exists()

    def run(self, *args, **kwargs):
        if self.path is not None:
            # only None when called in the __init__ function
            kwargs.setdefault('cwd', self.path)

        # NOTE if we do want to make a copy of environmental variables,
        # we must remove GIT_WORK_TREE
        kwargs['env'] = {}
        kwargs['stdout'] = PIPE
        kwargs['stderr'] = PIPE

        proc = Popen(args, **kwargs)
        (stdout, stderr) = proc.communicate()
        if proc.returncode != 0:
            raise CommandError(args[0], proc.returncode, stdout, stderr)
        return stdout

    def exists(self):
        """Check if the working directory exists

        Returns:
            bool - True if the working directory exists
        """
        return os.path.exists(self.path)

    def get_working_directory(self):
        """Get the working directory for this repo.

        Args:
            cls (class object): The class

        Returns:
            str - the path to the working directory

        Raises:
            CommandError
        """
        raise NotImplementedError  # pragma: no cover

    def install_hook(self, hook_name, hook_content):
        """Install the repository hook for this repo.

        Args:
            hook_name (str)
            hook_content (str)
        """
        raise NotImplementedError  # pragma: no cover

    def remove_ignored_files(self):
        """Remove files ignored by the repository
        """
        raise NotImplementedError  # pragma: no cover

    def remove_unstaged_files(self):
        """Remove all unstaged files. This does NOT remove ignored files.

        TODO this may be specific to git?
        """
        raise NotImplementedError  # pragma: no cover

    def clear(self, target_commit):
        """Resets the repository to the target commit, removing any staged,
        unstaged, and untracked files.

        Args:
            target_commit (str): the commit ID
        Raises:
            CommandError - if the commit does not exist
        """
        raise NotImplementedError  # pragma: no cover

    def private_dir(self):
        """Get the private directory associated with this repo, but untracked
        by the repo.

        Returns:
            str - absolute path
        """
        raise NotImplementedError  # pragma: no cover

    def repository_dir(self):
        """Get the directory used by the VCS to store repository info.

        e.g. .git for git

        Returns:
            str - absolute path
        """
        raise NotImplementedError  # pragma: no cover

    def get_signature(self):
        """Get the signature of the current state of the repository

        Returns:
            str
        """
        raise NotImplementedError  # pragma: no cover

    def ignore_patterns_file(self):
        """The ignore patterns file for this repo type.

        e.g. .gitignore for git

        Returns:
            str - file name
        """
        raise NotImplementedError  # pragma: no cover

    def path_is_ignored(self, path):
        """Given a path, check if the path would be ignored.

        Returns:
            boolean
        """
        raise NotImplementedError  # pragma: no cover

    def get_ignored_files(self):
        """Returns the list of files being ignored in this repository.

        Note that file names, not directories, are returned.

        So, we will get the following:

        a/b.txt
        a/c.txt

        instead of just:

        a/

        Returns:
            List[str] - list of ignored files. The paths are relative to the repo.
        """
        raise NotImplementedError  # pragma: no cover

    @contextmanager
    def temp_copy(self):
        """Yields a new Vcs object that represents a temporary, disposable
        copy of the current repository. The copy is deleted at the end
        of the context.

        The following are not copied:
        - ignored files
        - easyci private directory (.git/eci for git)

        Yields:
            Vcs
        """
        with contextmanagers.temp_dir() as temp_dir:
            temp_root_path = os.path.join(temp_dir, 'root')
            path = os.path.join(self.path, '')  # adds trailing slash
            check_call(['rsync', '-r', "--exclude={}".format(self.private_dir()), "--filter=dir-merge,- {}".format(
                self.ignore_patterns_file()), path, temp_root_path])
            copy = self.__class__(path=temp_root_path)
            yield copy



from setuptools import setup, find_packages

import easyci


tests_require = [
    'flake8>=2.4.0',
    'pytest>=2.5.0',
    'mock>=1.0.0',
    'coverage>=3.7.1',
]

install_requires = [
    'Click>=5.0',
    'enum34>=1.0.4',
    'filelock>=2.0.4',
    'PyYAML>=3.11',
    'subprocess32>=3.2.6',
    'watchdog>=0.8.2',
]

setup(
    name='easyci',
    version=easyci.__version__,
    author='Naphat Sanguansin',
    author_email='naphat.krit@gmail.com',
    description='Local CI, for mortals.',
    packages=find_packages(),
    package_data={
        'easyci.hooks': ['*'],
    },
    install_requires=install_requires,
    extras_require={'tests': tests_require},
    url='https://github.com/naphatkrit/easyci',
    download_url='https://github.com/naphatkrit/easyci/tarball/' + easyci.__version__,
    tests_require=tests_require,
    keywords=['continuous', 'integration', 'easy', 'ci',
              'gating', 'tests', 'testing', 'test', 'git'],
    entry_points='''
        [console_scripts]
        eci=easyci.cli:cli
    ''',
)



import hashlib
import os
import stat

from easyci.vcs.base import CommandError, Vcs


class GitVcs(Vcs):
    binary_path = 'git'

    def __init__(self, path=None):
        super(GitVcs, self).__init__(path)
        private_dir = self.private_dir()
        if not os.path.exists(private_dir):
            os.makedirs(private_dir)

    def get_working_directory(self):
        """Get the working directory for this repo.

        This is available as a class method as it is usually needed to
        initialize the VCS object itself.

        Args:
            cls (class object): The class

        Returns:
            str - the path to the working directory

        Raises:
            CommandError
        """
        return self.run('rev-parse', '--show-toplevel').strip()

    def run(self, *cmd, **kwargs):
        cmd = [self.binary_path] + list(cmd)
        return super(GitVcs, self).run(*cmd, **kwargs)

    def remove_ignored_files(self):
        """Remove files ignored by the repository
        """
        self.run('clean', '-fdX')

    def remove_unstaged_files(self):
        """Remove all unstaged files. This does NOT remove ignored files.

        TODO this may be specific to git?
        """
        self.run('clean', '-fd')  # remove untracked files
        self.run('checkout', self.path)  # revert changes to staged version

    def clear(self, target_commit):
        """Resets the repository to the target commit, removing any staged,
        unstaged, and untracked files.

        Args:
            target_commit (str): the commit ID
        Raises:
            CommandError - if the commit does not exist
        """
        self.run('reset', '--hard', target_commit)
        self.run('clean', '-fd')

    def private_dir(self):
        """Get the private directory associated with this repo, but untracked
        by the repo.

        Returns:
            str - absolute path
        """
        return os.path.join(self.repository_dir(), 'eci')

    def repository_dir(self):
        """Get the directory used by the VCS to store repository info.

        e.g. .git for git

        Returns:
            str - absolute path
        """
        return os.path.join(self.path, '.git')

    def get_signature(self, base_commit=None):
        """Get the signature of the current state of the repository

        TODO right now `get_signature` is an effectful process in that
        it adds all untracked file to staging. This is the only way to get
        accruate diff on new files. This is ok because we only use it on a
        disposable copy of the repo.

        Args:
            base_commit - the base commit ('HEAD', sha, etc.)

        Returns:
            str
        """
        if base_commit is None:
            base_commit = 'HEAD'
        self.run('add', '-A', self.path)
        sha = self.run('rev-parse', '--verify', base_commit).strip()
        diff = self.run('diff', sha).strip()
        if len(diff) == 0:
            try:
                return self.get_signature(base_commit + '~1')
            except CommandError:
                pass
        h = hashlib.sha1()
        h.update(sha)
        h.update(diff)
        return h.hexdigest()

    def install_hook(self, hook_name, hook_content):
        """Install the repository hook for this repo.

        Args:
            hook_name (str)
            hook_content (str)
        """
        hook_path = os.path.join(self.path, '.git/hooks', hook_name)
        with open(hook_path, 'w') as f:
            f.write(hook_content)
        os.chmod(hook_path, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE)

    def get_ignored_files(self):
        """Returns the list of files being ignored in this repository.

        Note that file names, not directories, are returned.

        So, we will get the following:

        a/b.txt
        a/c.txt

        instead of just:

        a/

        Returns:
            List[str] - list of ignored files. The paths are absolute.
        """
        return [os.path.join(self.path, p) for p in
                self.run('ls-files', '--ignored', '--exclude-standard',
                         '--others').strip().split()
                ]

    def ignore_patterns_file(self):
        """The ignore patterns file for this repo type.

        e.g. .gitignore for git

        Returns:
            str - file name
        """
        return '.gitignore'

    def path_is_ignored(self, path):
        """Given a path, check if the path would be ignored.

        Returns:
            boolean
        """
        try:
            self.run('check-ignore', '--quiet', path)
        except CommandError as e:
            if e.retcode == 1:
                # path is ignored
                return False
            else:
                # fatal error
                raise e
        return True



import click

import easyci

from easyci.commands.clear_history import clear_history
from easyci.commands.gc import gc
from easyci.commands.init import init
from easyci.commands.test import test
from easyci.commands.watch import watch
from easyci.vcs.git import GitVcs
from easyci.version import get_installed_version, VersionNotInstalledError


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=easyci.__version__, prog_name='EasyCI')
@click.pass_context
def cli(ctx):
    git = GitVcs()
    if ctx.args[0] != 'init':
        try:
            version = get_installed_version(git)
        except VersionNotInstalledError:
            click.echo('Please run `eci init` first.')
            ctx.abort()
        if version != easyci.__version__:
            click.echo('EasyCI version mismatch. Please rerun `eci init`.')
            ctx.abort()
    ctx.obj = dict()
    ctx.obj['vcs'] = git

cli.add_command(clear_history)
cli.add_command(gc)
cli.add_command(init)
cli.add_command(test)
cli.add_command(watch)



import click

from easyci.history import clear_history as clear_history_internal


@click.command('clear-history')
@click.pass_context
def clear_history(ctx):
    """Clear tests run history. History is normally used to keep track of
    whether a test has been run for a specific state of the project, to
    avoid running tests redundantly. This command clears the history,
    causing the next `eci test` command to always run tests.
    """
    git = ctx.obj['vcs']
    clear_history_internal(git)



import click

from easyci import locking
from easyci.history import get_committed_signatures, get_staged_signatures
from easyci.results import get_signatures_with_results, remove_results


@click.command('gc')
@click.pass_context
def gc(ctx):
    """Runs housekeeping tasks to free up space.

    For now, this only removes saved but unused (unreachable) test results.
    """
    vcs = ctx.obj['vcs']
    count = 0
    with locking.lock(vcs, locking.Lock.tests_history):
        known_signatures = set(get_committed_signatures(vcs) + get_staged_signatures(vcs))
        for signature in get_signatures_with_results(vcs):
            if signature not in known_signatures:
                count += 1
                remove_results(vcs, signature)
    click.echo('Removed {} saved results.'.format(count))
